# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2014, 2015 Cumulus Networks, Inc. All rights reserved.
# Copyright (C) 2014 Metacloud Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.
from __future__ import absolute_import

import atexit
import logging
import os
import signal
import socket
import subprocess
import sys

from abc import ABCMeta, abstractmethod
import greenlet
import eventlet
from daemon import daemon

from netq.common import mgmt
from netq.common import utils
from netq.common.enums import LogDestination


class Service(object):
    """ Abstract base class.
    """
    __metaclass__ = ABCMeta

    def __init__(self, conf, proc_name):
        self._conf = conf
        # Set up signal handlers before daemonizing.
        self._run_gt = None
        self.__setup_signal_handlers()
        self.__daemon_context = daemon.DaemonContext(
            working_directory='/var/run'
        )
        if self._conf.daemon:
            self.__daemon_context.open()
        # Patch system modules to be greenthread-friendly. Must be after
        # daemonizing.
        eventlet.monkey_patch()
        self._pool = eventlet.GreenPool(size=self._conf.concurrency)
        # Set up logging. Must be after daemonizing.

        if self._conf.debug:
            lvl = logging.DEBUG
        else:
            lvl = getattr(logging, self._conf.loglevel)
        if self._conf.logdest in {LogDestination.SYSLOG,
                                  LogDestination.STDOUT}:
            self._logger = utils.get_logger(proc_name,
                                            self._conf.logdest)
        else:
            self._logger = utils.get_logger(
                proc_name,
                LogDestination.LOGFILE,
                filehandler_args={'filename': self._conf.logdest,
                                  'maxBytes': self._conf.logfilesize,
                                  'backupCount': self._conf.logbackupcount}
            )
        self._logger.setLevel(lvl)
        self._mgmtserver = mgmt.MgmtServer(self._conf.udsfile,
                                           self._process,
                                           self._logger)
        self.__pidfd = utils.Pidfile(self._conf.pidfile,
                                     proc_name)

    @abstractmethod
    def _process(self, msg):
        """ Process requests from a mgmt. client.
        Must return a response object and an Exception object. The
        latter is None if no exception.
        Over-ride this method in the derived class
        """
        raise NotImplementedError

    @abstractmethod
    def _run(self):
        """ Run method to be implemented by the derived class.
        """
        raise NotImplementedError

    def _serve(self, sock, handle, pool=None, bufsize=None, err_cbs=None):
        """ Runs a udp listener on the supplied socket. Calls the function
        *handle* in a separate greenthread for every incoming datagram.
        *handle* takes two arguments: packet buffer and client address.
        """
        # pylint: disable=too-many-arguments
        err_cbs = err_cbs or {}
        bufsize = bufsize or self._conf.max_packet_size
        pool = pool or self._pool
        while True:
            try:
                buf, addr = sock.recvfrom(bufsize)
                green_thread = pool.spawn(handle, buf, addr)
                green_thread.link(self._stop_checker)
            except socket.error as (err, _):
                if err in err_cbs:
                    err_cbs[err]()
                    return
                else:
                    raise
            except eventlet.StopServe:
                return

    def _serve_command(self, cmd, handle, pool=None):
        """ FIXME: Runs a tcp server on the supplied socket. Calls the
        function *handle* in a separate greenthread for every incoming client
        connection. *handle* takes two arguments: the client socket object,
        and the client address.
        """
        pool = pool or self._pool
        bpm = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
        while True:
            try:
                line = bpm.stdout.readline()
                if not line:
                    raise Exception('Command %s terminated unexpectedly' % cmd)
                green_thread = pool.spawn(handle, line)
                green_thread.link(self._stop_checker)
            except eventlet.StopServe:
                return
            finally:
                try:
                    bpm.kill()
                except Exception:  # pylint: disable=broad-except
                    pass

    def _serve_tcp(self, sock, handle, pool=None):
        """ Runs a tcp server on the supplied socket. Calls the function
        *handle* in a separate greenthread for every incoming client
        connection. *handle* takes two arguments: the client socket object,
        and the client address.
        """
        pool = pool or self._pool
        while True:
            try:
                client_sock, addr = sock.accept()
                green_thread = pool.spawn(handle, client_sock, addr)
                green_thread.link(self._stop_checker)
            except eventlet.StopServe:
                return

    def _stop_checker(self, green_thread):
        """ Propagates exceptions to the run green thread.
        """
        try:
            green_thread.wait()
        except greenlet.GreenletExit:  # pylint: disable=no-member
            pass
        except Exception:  # pylint: disable=broad-except
            eventlet.kill(self._run_gt, *sys.exc_info())

    def __delpid(self):
        """ Removes the pidfile when the process exits gracefully.
        """
        try:
            os.remove(self._conf.pidfile)
        except OSError:
            self._logger.exception('Unable to remove pid file on exit')

    def __setup_signal_handlers(self):
        """ Set up signal handlers.
        """
        signal.signal(signal.SIGINT, self.__term_handler)
        signal.signal(signal.SIGTERM, self.__term_handler)
        signal.signal(signal.SIGHUP, self.__term_handler)

    def __term_handler(self, signum, _):
        """ Signal handler that terminates the process.
        """
        msg = 'Caught signal %d' % signum
        self._logger.debug(msg)
        # _run_gt is initialized at init but stuffed in run().
        # But we could get here before that from a SIG.
        if self._run_gt is not None:
            eventlet.kill(self._run_gt, SystemExit(msg))
        else:
            raise SystemExit(msg)

    def get_logger(self):
        """ Returns the logger object for the instance.
        """
        return self._logger

    def run(self):
        """ Once upon a time...
        """
        try:
            # Make sure that I am not already running.
            pid = os.getpid()
            self.__pidfd.lock()
            self.__pidfd.write(pid)
            atexit.register(self.__delpid)
            self._logger.info('Starting (pid %d) ...', pid)
            atexit.register(self._logger.info,
                            'Terminating (pid %d)' % pid)
            self._pool.spawn(self._mgmtserver.run).link(self._stop_checker)
            self._run_gt = self._pool.spawn(self._run)
            self._run_gt.wait()
        except SystemExit:
            pass
        except:  # pylint: disable=broad-except
            self._logger.exception('Error: ')
            raise
        finally:
            self.__daemon_context.close()
